#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import asyncio

import re

from typing import Type, Union, List

from ...models import Artist, Album, Track, Playlist
from ...config import VK, headers, playlistsPerReq, playlistsOwnerId


class Get:
    async def get(self, ownerId: int, trackId: int, includeLyrics: bool = False) -> Union[Track, dict]:
        """
        Получает информацию об аудиотреке по его идентификатору.

        Пример использования:\n
        result = await client.get(ownerId=474499244, trackId=456638035, includeLyrics=True)\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int)
        :param trackId: идентификатор аудиотрека, информацию о котором необходимо получить. (int)
        :param includeLyrics: флаг, указывающий, необходимо ли включать текст трека в ответ. (bool, по умолчанию `False`)
        :return: информация об аудиотреке в виде объекта модели `Track` или словаря.
        """

        id = f"{ownerId}_{trackId}"

        tasks = [self._VKReq("getById", {"audios": id})]

        if includeLyrics:
            tasks.append(self._VKReq("getLyrics", {"audio_id": id}))

        responses = await asyncio.gather(*tasks)

        track = responses[0]
        if not track:
            return self._raiseError("trackNotFound")

        if includeLyrics:
            lyrics = responses[1]
            if not lyrics.get("error"):
                lyrics = lyrics.get("lyrics")

                timestamps = lyrics.get("timestamps")
                if timestamps:
                    lyrics = [line.get("line") for line in timestamps]

                else:
                    lyrics = lyrics.get("text")

                track["lyrics"] = "\n".join(lyrics)

        return self._finalizeResponse(track, Track)


    async def getTracks(self, groupId: int = None) -> Union[List[Union[Track, dict]], Track, dict, None]:
        """
        Получает треки пользователя или группы по его (её) идентификатору. (Временно не работает)

        Пример использования:\n
        result = await client.getTracks(groupId=-215973356)\n
        print(result)

        :param groupId: идентификатор пользователя или группы. (int, по умолчанию текущий пользователь)
        :return: список аудиотреков в виде объектов класса `Track` или словарей, аудиотрек в виде объекта модели `Track` (если он единственный), или `None` (если треки отсутствуют).
        """

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        tracks = await self._client.sendReq(VK + "audios" + str(groupId), cookies=self._cookies if hasattr(self, "_cookies") else None, headers=headers, responseType="code")
        tracks = await self._getTracks(tracks)

        return tracks


    async def getArtist(self, artistId: int, includeAlbums: bool = False, includeTracks: bool = False) -> Union[Artist, dict]:
        """
        Получает информацию об артисте по его идентификатору.

        Пример использования:\n
        result = await client.getArtist(artistId=5696274288194638935, includeAlbums=True, includeTracks=True)\n
        print(result)

        :param artistId: идентификатор артиста, информацию о котором необходимо получить. (int)
        :param includeAlbums: флаг, указывающий, необходимо ли включать альбомы артиста в ответ. (bool, по умолчанию `False`)
        :param includeTracks: флаг, указывающий, необходимо ли включать треки артиста в ответ. (bool, умолчанию `False`)
        :return: информация об артисте в виде объекта модели `Artist` или словаря.
        """

        params = {"artist_id": artistId}

        tasks = [self._VKReq("getArtistById", params)]

        if includeAlbums:
            tasks.append(self._VKReq("getAlbumsByArtist", params))

        if includeTracks:
            tasks.append(self._VKReq("getAudiosByArtist", params))

        responses = await asyncio.gather(*tasks)

        artist = responses[0]
        if not artist.get("name"):
            return self._raiseError("artistNotFound")

        if includeAlbums:
            artist["albums"] = responses[1].get("items")

        if includeTracks:
            artist["tracks"] = responses[2].get("items")

        return self._finalizeResponse(artist, Artist)


    async def getRelatedArtists(self, artistId: int, limit: int = 10) -> Union[List[Union[Artist, dict]], Artist, dict, None]:
        """
        Получает похожих артистов.

        Пример использования:\n
        result = await client.getRelatedArtists(artistId=5696274288194638935, limit=5)\n
        print(result)

        :param artistId: идентификатор артиста, похожих на которого необходимо получить. (int)
        :param limit: максимальное количество артистов, которое необходимо вернуть. (bool, по умолчанию 10)
        :return: список артистов в виде объектов модели `Artist` или словарей, артист в виде объекта модели `Artist` или словаря (если он единственственный), или None (если `artistId` неверный или похожие артисты отсутствуют).
        """

        return self._finalizeResponse((await self._VKReq("getRelatedArtistsById", {"artist_id": artistId, "count": limit})).get("artists"), Artist)


    async def _getTracks(self, tracks: str, objectType: Union[Type[Union[Album, Playlist]], None] = None) -> dict:
        tracks = re.sub(r"\\/", "/", re.sub(r"false", "False", re.sub(r"true", "True", tracks)))
        try:
            if objectType:
                tracks = eval(tracks[tracks.rfind("[["): tracks.rfind("]]") + 2])

            else:
                tracks = []

        except SyntaxError:
            tracks = self._raiseError("accessDenied" + ("WithoutCookie" if not hasattr(self, "_cookies") else ""))

        if isinstance(tracks, list):
            for index, track in enumerate(tracks):
                tracks[index] = Track({"owner_id": track[1], "id": track[0], "title": track[3], "subtitle": track[16], "main_artists": track[17], "duration": track[5]}) if len(track) > 3 else None

            if not tracks or all(track is None for track in tracks):
                tracks = None

        return tracks


    async def getAlbum(self, ownerId: int, albumId: int, includeTracks: bool = False) -> Union[Album, dict]:
        """
        Получает информацию об альбоме по его идентификатору.

        Пример использования:\n
        result = await client.getAlbum(ownerId=-2000837600, albumId=16837600, includeTracks=True)\n
        print(result)

        :param ownerId: идентификатор владельца альбома (пользователь или группа). (int)
        :param albumId: идентификатор альбома, информацию о котором необходимо получить. (int)
        :param includeTracks: флаг, указывающий, необходимо ли включать треки альбома в ответ. (bool, по умолчанию `False`)
        :return: информация об альбоме в виде объекта модели `Album` или словаря.
        """

        tasks = [self._VKReq("getPlaylistById", {"owner_id": ownerId, "playlist_id": albumId})]

        if includeTracks:
            tasks.append(self._client.sendReq(VK + "music/album/" + f"{ownerId}_{albumId}", headers=headers, responseType="code"))

        responses = await asyncio.gather(*tasks)

        album = responses[0]
        if not album:
            return self._raiseError("albumNotFound")

        if includeTracks:
            album["tracks"] = await self._getTracks(responses[1], Album)

        return self._finalizeResponse(album, Album)


    async def getPlaylist(self, playlistId: int, ownerId: int = None, includeTracks: bool = False) -> Union[Playlist, dict]:
        """
        Получает информацию о плейлисте по его ID.

        Пример использования:\n
        result = await client.getPlaylist(playlistId=1, ownerId=-215973356, includeTracks=True)\n
        print(result)

        :param playlistId: идентификатор плейлиста, информацию о котором необходимо получить. (int)
        :param ownerId: идентификатор владельца плейлиста (пользователь или группа). (int, по умолчанию текущий пользователь)
        :param includeTracks: флаг, указывающий, необходимо ли включать треки плейлиста в ответ. (bool, по умолчанию `False`)
        :return: информация о плейлисте в виде объекта модели `Playlist` или словаря.
        """

        if not ownerId:
            ownerId = (await self.getSelf()).get("id")

        tasks = [self._VKReq("getPlaylistById", {"owner_id": ownerId, "playlist_id": playlistId})]

        if includeTracks:
            tasks.append(self._client.sendReq(VK + "music/playlist/" + f"{ownerId}_{playlistId}", cookies=self._cookies if hasattr(self, "_cookies") else None, headers=headers, responseType="code"))

        responses = await asyncio.gather(*tasks)

        playlist = responses[0]
        if playlist.get("error"):
            return playlist

        if includeTracks:
            playlist["tracks"] = await self._getTracks(responses[1], Playlist)
        return self._finalizeResponse(playlist, Playlist)


    async def getPlaylists(self, ownerId: int = None, playlistTypes: Union[str, List[str]] = ["own", "foreign", "album"]) -> Union[List[Union[Playlist, Album, dict]], Playlist, Album, dict, None]:
        """
        Получает плейлисты пользователя или группы.

        Пример использования:\n
        result = await client.getPlaylists(ownerId=-215973356, playlistTypes="own")\n
        print(result)

        :param ownerId: идентификатор пользователя или группы, плейлисты которого(ой) необходимо получить. (int, по умолчанию текущий пользователь)
        :param playlistTypes: типы плейлистов, которые необходимо получить: `own` — принадлежащий пользователю или группе, `foreign` — не принадлежащий пользователю или группе, `album` — альбом. (str или list, по умолчанию ["own", "foreign", "album"])
        :return: список плейлистов в виде объектов модели `Playlist` или `Album` или словарей, плейлист в виде объекта модели `Playlist` или `Album` или словаря (если он единственственный), или `None` (если плейлисты отсутствуют).
        """

        if not ownerId:
            ownerId = (await self.getSelf()).get("id")

        if not isinstance(playlistTypes, list):
            playlistTypes = [playlistTypes]

        for index, playlistType in enumerate(playlistTypes):
            playlistTypes[index] = playlistType.lower()

        method, params = "getPlaylists", {"owner_id": ownerId, "count": playlistsPerReq}
        playlists_ = await self._VKReq(method, params)
        if playlists_.get("error"):
            return playlists_

        playlists = [playlist for playlist in playlists_.get("items")]
        count = playlists_.get("count")
        offset = count if count < playlistsPerReq else playlistsPerReq

        if offset < count:
            tasks = []
            while offset < count:
                tasks.append(self._VKReq(method, {**params, **{"offset": offset}}))
                offset += playlistsPerReq

            playlists_ = await asyncio.gather(*tasks)
            for playlistGroup in playlists_:
                for playlist in playlistGroup.get("items"):
                    playlists.append(playlist)

        playlists = self._finalizeResponse(playlists, Playlist)

        if not isinstance(playlists, list):
            playlists = [playlists]

        playlists = [playlist for playlist in playlists if (isinstance(playlist, dict) and not playlist.get("artists") and ("own" if playlist.get("own") else "foreign") in playlistTypes) or (isinstance(playlist, Playlist) and ("owm" if playlist.own else "foreign") in playlistTypes) or ((isinstance(playlist, Album) or (isinstance(playlist, dict) and playlist.get("artists"))) and "album" in playlistTypes)]

        return (playlists if len(playlists) > 1 else playlists[0]) if playlists else None


    async def getCuratorTracks(self, curatorId: int, limit: int = 10, offset: int = 0) -> Union[List[Union[Track, dict]], Track, dict, None]:
        """
        Получает аудиотреки, принадлежащие куратору.

        Пример использования:\n
        result = await client.getCuratorTracks(curatorId=28905875, limit=5)\n
        print(result)

        :param curatorId: идентификатор куратора (пользователь или группа). (int)
        :param limit: максимальное количество аудиотреков, которое необходимо вернуть. (int, по умолчанию 10)
        :param offset: количество результатов, которые необходимо пропустить. (int, необязательно)
        :return: список аудиотреков в виде объектов модели Track или словарей, аудиотрек в виде объекта модели `Track` или словаря (если он единственный), или `None` (если неверный `curatorId` или аудиотреки отсутствуют).
        """

        return self._finalizeResponse((await self._VKReq("getAudiosByCurator", {"curator_id": curatorId, "count": limit, "offset": offset})).get("items"), Track)


    async def getTracksFromFeed(self) -> List[dict]:
        """
        Получает все треки из новостной ленты.

        Пример использования:\n
        result = await client.getTracksFromFeed()\n
        print(result)

        :return: список аудиотреков в виде словарей с ключами `ownerId` и `trackId`.
        """

        tracks = (await self._VKReq("getAudioIdsBySource", {"source": "feed"})).get("audios")
        for index, track in enumerate(tracks):
            ownerId, trackId = track.get("audio_id").split("_")[:2]
            tracks[index] = {"ownerId": int(ownerId), "trackId": int(trackId)}

        return tracks


    async def getRecommendations(self, limit: int = 10, offset: int = 0, ownerId: int = None, trackId: int = None) -> Union[List[Union[Track, dict]], Track, dict, None]:
        """
        Получает рекомендации аудиотреков для пользователя или похожие на аудиотрек.

        Пример использования для рекомендаций пользователя:\n
        result = await client.getRecommendations(limit=20)\n
        print(result)

        Пример использования для рекомендаций по аудиотреку:\n
        result = await client.getRecommendations(limit=5, ownerId=474499156, trackId=456637846)\n
        print(result)

        :param limit: максимальное количество аудиотреков, которое необходимо вернуть. (int, по умолчанию 10, минимально для пользовательских рекомендаций 10)
        :param offset: количество результатов, которые необходимо пропустить. (int, необязательно)
        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа).
        :param trackId: идентификатор аудиотрека, похожие на который необходимо получить.
        :return: список аудиотреков в виде объектов модели `Track` или словарей, аудиотрек в виде объекта модели `Track` или словаря (если он единственный), или `None` (если рекомендации или похожие треки отсутствуют).
        """

        if not all((ownerId, trackId)) and limit < 10:
            limit = 10

        return self._finalizeResponse((await self._VKReq("getRecommendations", {"count": limit, "offset": offset, **({"target_audio": f"{ownerId}_{trackId}"} if all((ownerId, trackId)) else {})})).get("items"), Track)


    async def getNew(self) -> List[Union[Track, dict]]:
        """
        Получает аудиотреки, вышедшие недавно.

        Пример использования:\n
        result = await client.getNew()\n
        print(result)

        :return: список аудиотреков в виде объектов модели `Track` или словарей.
        """

        playlist = await self.getPlaylist(2, playlistsOwnerId, True)

        return playlist.tracks if isinstance(playlist, Playlist) else playlist.get("tracks")


    async def getPopular(self) -> List[Union[Track, dict]]:
        """
        Получает популярные аудиотреки.

        Пример использования:\n
        result = await client.getPopular()\n
        print(result)

        :return: список аудиотреков в виде объектов модели `Track` или словарей.
        """

        playlist = await self.getPlaylist(1, playlistsOwnerId, True)

        return playlist.tracks if isinstance(playlist, Playlist) else playlist.get("tracks")


    async def getEditorsPicks(self) -> List[Union[Track, dict]]:
        """
        Получает аудиотреки, выбранные редакторами ВКонтакте.

        Пример использования:\n
        result = await client.getEditorsPicks()\n
        print(result)

        :return: список аудиотреков в виде объектов модели `Track` или словарей.
        """

        playlist = await self.getPlaylist(3, playlistsOwnerId, True)

        return playlist.tracks if isinstance(playlist, Playlist) else playlist.get("tracks")


    async def getTrackCount(self, ownerId: int) -> int:
        """
        Получает количество аудиотреков, принадлежащих этому пользователю или группе.

        Пример использования:\n
        result = await client.getTrackCount(ownerId=-215973356)\n
        print(result)

        :param ownerId: идентификатор пользователя или группы. (int)
        :return: количество аудиотреков, принадлежащих пользователю или группе, в виде целого числа.
        """

        return await self._VKReq("getCount", {"owner_id": ownerId})


    async def getSearchTrends(self, limit: int = 10, offset: int = 0) -> Union[List[str], str, None]:
        """
        Получает самые частые поисковые запросы в музыке.

        Пример использования:\n
        result = await client.getSearchTrends(limit=5)\n
        print(result)

        :param limit: максимальное количество запросов, которое необходимо вернуть. (int, по умолчанию 10)
        :param offset: количество результатов, которые необходимо пропустить. (int, необязательно)
        :return: список строк, представляющих самые частые поисковые запросы в музыке, строка, представляющая запрос (если он единственный) или `None` (если запросы отсутствуют).
        """

        return [item.get("name") for item in (await self._VKReq("getSearchTrends", {"count": limit, "offset": offset})).get("items")]


    async def getBroadcast(self, id: int = None, isGroup: bool = False) -> Union[Track, dict, None, bool]:
        """
        Получает аудиотрек, транслируемый в статус.

        Пример использования для текущего пользователя:\n
        result = await client.getBroadcast()\n
        print(result)

        Пример использования для любого пользователя:\n
        result = await client.getBroadcast(id=1)\n
        print(result)

        Пример использования для группы:\n
        result = await client.getBroadcast(id=-215973356, isGroup=True)\n
        print(result)

        :param id: идентификатор пользователя или группы, аудиотрек из статуса которого(ой) необходимо получить. (int, по умолчанию текущий пользователь)
        :param isGroup: флаг, указывающий необходимо получить статус пользователя или группы (`True` для группы). (bool, по умолчанию `False`)
        :return: аудиотрек в виде объекта модели `Track` или словаря, `None` (если ничего не проигрывается), или `False` (если музыка не транслируется в статус, только для текущего пользователя).
        """

        if isGroup and id:
            id = id - (id * 2)

        broadcast = await self._VKReq("status.get", {"user_id": id} if id else None)
        if broadcast.get("error"):
            return broadcast

        audio = broadcast.get("audio")
        if audio:
            return await self.get(audio.get("owner_id"), audio.get("id"), True)

        else:
            if not id or id == (await self.getSelf()).get("id"):
                isBroadcastEnabled = (await self._VKReq("getBroadcast")).get("enabled")
                if not bool(isBroadcastEnabled):
                    return False

            return None


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import aiofiles

import os

import pytz
from datetime import datetime

from typing import Union, Optional, List, Tuple

from ...models import Track, Playlist
from ...config import moscowTz
from ...utils.another import checkFile


class Update:
    async def upload(self, filename: str, title: str = None, artist: str = None, lyrics: str = None, genreId: int = None, removeFromSearchResults: bool = None, groupId: int = None) -> Union[Track, dict]:
        """
        Загружает новый аудиотрек во ВКонтакте.

        Пример загрузки файла с названием «prombl — zapreti.mp3»:\n
        result = await client.upload(filename="prombl — zapreti", title="zapreti", artist="prombl", lyrics="yourLyrics", removeFromSearchResults=True, groupId="yourGroupId")\n
        print(result)

        :param filename: имя MP3-файла, содержащего аудиотрек, который необходимо загрузить (без расширения). (str)
        :param title: название аудиотрека. (str, необязательно)
        :param artist: артист(ы) аудиотрека. (str, необязательно)
        :param lyrics: текст аудиотрека. (str, необязательно, временно не работает)
        :param genreId: жанр аудиотрека (в виде идентификатора). (int, необязательно, временно не работает)
        :param removeFromSearchResults: флаг, указывающий, будет ли аудиотрек скрыт из поисковой выдачи. (bool, необязательно, временно не работает)
        :param groupId: идентификатор группы, в музыку которой необходимо загрузить аудиотрек (int, необязательно, временно не работает)
        :return: загруженный аудиотрек в виде объекта модели `Track` или словаря.
        """

        if filename.endswith(".mp3"):
            filename = filename[:-4]

        filename = checkFile(filename + ".mp3")
        if not filename:
            self._raiseError("MP3FileNotFound")

        fileSizeInBytes = os.path.getsize(filename)
        fileSizeInMB = fileSizeInBytes / (1024 * 1024)  # Преобразование размера файла в мегабайты

        if fileSizeInMB > 200:  # Проверка, меньше ли размер файла 200 MB
            self._raiseError("MP3FileTooBig")

        uploadUrl = (await self._VKReq("getUploadServer")).get("upload_url")

        async with aiofiles.open(filename, "rb") as file:
            fileContent = await file.read()

        uploadingFileResponse = await self._client.sendReq(uploadUrl, files={"file": (filename, fileContent, "audio/mpeg")}, method="POST")

        server = uploadingFileResponse.get("server")
        audio = uploadingFileResponse.get("audio")
        hash = uploadingFileResponse.get("hash")

        response = await self._VKReq("save", {"server": server, "audio": audio, "hash": hash, "title": title, "artist": artist})

        return self._finalizeResponse(response, Track) if not response.get("error") else response


    async def add(self, ownerIds: Union[int, List[int]], trackIds: Union[int, List[int]], playlistId: Optional[int] = None, groupId: Optional[int] = None) -> Tuple[Union[bool, dict]]:
        """
        Добавляет аудиотрек в музыку или плейлист пользователя или группы.

        Пример использования:\n
        result = await client.add(ownerIds=474499244, trackIds=456638035, playlistId="yourPlaylistId", groupId="yourGroupId")\n
        print(result)

        :param ownerIds: идентификатор(ы) владельца аудиотрека(ов) (пользователь или группа). (int или list)
        :param trackIds: идентификатор(ы) аудиотрека(ов), который(е) необходимо добавить. (int или list)
        :param playlistId: идентификатор плейлиста, в который необходимо добавить аудиотрек. (int, необязательно)
        :param groupId: идентификатор группы, в музыку или плейлист которой необходимо добавить аудиотрек. (int, необязательно)
        :return: `True`, если аудиотрек успешно добавлен, `False` в противном случае.
        """

        if type(ownerIds) != type(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        if isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not (isinstance(ownerIds, list) and isinstance(trackIds, list)):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        if playlistId:
            method = "addToPlaylist"

        else:
            method = "add"

        results = []
        for ownerId, trackId in zip(ownerIds, trackIds):
            response = await self._VKReq(method, {**({"owner_id": ownerId, "audio_id": trackId, "group_id": groupId} if not playlistId else {"owner_id": groupId, "audio_ids": f"{ownerId}_{trackId}"}), "playlist_id": playlistId})
            results.append(bool(response) if not (isinstance(response, dict) and response.get("error")) else self._raiseError("playlistNotFound"))

        return tuple(results)


    async def remove(self, ownerIds: Union[int, List[int]], trackIds: Union[int, List[int]], playlistId: Optional[int] = None, groupId: Optional[int] = None, reValidateIds: bool = True) -> Tuple[Union[bool, dict]]:
        """
        Удаляет аудиотрек из музыки или плейлиста пользователя или группы.

        Пример использования:\n
        result = await client.remove(ownerIds=474499244, trackIds=456638035, playlistId="yourPlaylistId", groupId="yourGroupId", reValidateIds=False)\n
        print(result)

        :param ownerIds: идентификатор(ы) владельца аудиотрека(ов) (пользователь или группа). (int или list)
        :param trackIds: идентификатор(ы) аудиотрека(ов), который(е) необходимо удалить. (int или list)
        :param playlistId: идентификатор плейлиста, из которого необходимо удалить аудиотрек(и). (int, необязательно, метод временно не работает для плейлистов, привязанных к чату)
        :param groupId: идентификатор группы, из музыки или плейлиста которой необходимо удалить аудиотрек(и). (int, необязательно)
        :param reValidateIds: флаг, указывающий, необходимо ли перепроверить идентификатор(ы) аудитрека(ов) по находящимся в плейлисте. (bool, по умолчанию `True`)
        :return: `True`, если аудиотрек(и) успешно удалён, `False` в противном случае.
        """

        if type(ownerIds) != type(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        if isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            return self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        if not (isinstance(ownerIds, list) and isinstance(trackIds, list)):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        if reValidateIds:
            if playlistId:
                playlist = await self.getPlaylist(playlistId, groupId, includeTracks=True)
                if isinstance(playlist, dict) and playlist.get("error"):
                    return playlist

                if isinstance(playlist, dict):
                    existTracks = playlist.get("tracks")

                elif isinstance(playlist, Playlist):
                    existTracks = playlist.tracks if hasattr(playlist, "tracks") else None

            else:
                existTracks = await self.getTracks(groupId)

            if not existTracks and playlistId:
                return (False,)

            for index, (ownerId, trackId) in enumerate(zip(ownerIds, trackIds)):
                if not playlistId:
                    continue

                track = await self.get(ownerId, trackId)

                if isinstance(track, dict):
                    if track.get("error"):
                        continue

                    trackTitle = track.get("title")
                    trackArtists = track.get("artists")
                    if not trackArtists:
                        trackArtists = track.get("artist")

                elif isinstance(track, Track):
                    trackTitle = track.title
                    trackArtists = track.artists
                    if not trackArtists:
                        trackArtists = track.artist

                for existTrack in existTracks:
                    if isinstance(existTrack, dict):
                        existTrackTitle = existTrack.get("title")
                        existTrackArtists = existTrack.get("artists")
                        if not existTrackArtists:
                            existTrackArtists = existTrack.get("artist")

                        existTrackOwnerId = existTrack.get("ownerId")
                        existTrackId = existTrack.get("trackId")

                    elif isinstance(existTrack, Track):
                        existTrackTitle = existTrack.title
                        existTrackArtists = existTrack.artists
                        if not existTrackArtists:
                            existTrackArtists = existTrack.artist

                        existTrackOwnerId = existTrack.ownerId
                        existTrackId = existTrack.trackId

                    if existTrackTitle == trackTitle and existTrackArtists == trackArtists:
                        ownerIds[index] = existTrackOwnerId
                        trackIds[index] = existTrackId
                        break

        deleteSemaphore = asyncio.Semaphore(5)
        if playlistId:
            if groupId > 0:
                groupId = -groupId

            async def removeTrackFromPlaylist(id: str, semaphore: asyncio.Semaphore) -> bool:
                async with semaphore:
                    response = await self._VKReq("removeFromPlaylist", {"audio_ids": id, "owner_id": groupId, "playlist_id": playlistId})
                    return bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            tasks = [removeTrackFromPlaylist(f"{ownerId}_{trackId}", deleteSemaphore) for ownerId, trackId in zip(ownerIds, trackIds)]

        else:
            async def removeTrack(ownerId: int, trackId: int, semaphore: asyncio.Semaphore) -> bool:
                async with semaphore:
                    response = await self._VKReq("delete", {"owner_id": ownerId, "audio_id": trackId, "group_id": groupId})
                    return bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            tasks = [removeTrack(ownerId, trackId, deleteSemaphore) for ownerId, trackId in zip(ownerIds, trackIds)]

        return tuple(await asyncio.gather(*tasks))


    async def edit(self, ownerId: int, trackId: int, title: Union[str, int] = None, artist: Union[str, int] = None, lyrics: Union[str, int] = None, genreId: int = None, removeFromSearchResults: bool = None) -> Union[bool, dict]:
        """
        Изменяет информацию об аудиотреке.

        Пример использования:\n
        result = await client.edit(ownerId="yourOwnerId", trackId="yourTrackId", title="zapreti", artist="prombl", "lyrics"="", "genreId"=3, removeFromSearchResults=True)\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int)
        :param trackId: идентификатор аудиотрека, информацию которого необходимо изменить. (int)
        :param title: новое название аудиотрека. (str, необязательно)
        :param artist: новый(е) артист(ы) аудиотрека. (str, необязательно)
        :param lyrics: новый текст аудиотрека. (str, необязательно)
        :param genreId: новый жанр аудиотрека (в виде идентификатора). (int, необязательно)
        :param removeFromSearchResults: флаг, указывающий, будет ли аудиотрек скрыт из поисковой выдачи. (bool, необязательно)
        :return: `True`, если информация аудиотрека успешно обновлена, `False` в противном случае.
        """

        if not any((title, artist, lyrics is not None, genreId is not None, removeFromSearchResults is not None)):
            return False

        response = await self._VKReq("edit", {"owner_id": ownerId, "audio_id": trackId,
            **({"title": title} if title else {}),
            **({"artist": artist} if artist else {}),
            **({"lyrics": lyrics} if lyrics is not None else {}),
            **({"genre_id": genreId} if genreId is not None else {}),
            **({"no_search": removeFromSearchResults} if removeFromSearchResults is not None else {})
        })

        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def restore(self, ownerId: int, trackId: int) -> Union[bool, dict]:
        """
        Восстанавливает удалённый аудиотрек.

        Пример использования:\n
        result = await client.restore(ownerId="yourOwnerId", trackId="yourTrackId")\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int)
        :param trackId: идентификатор аудиотрека, который необходимо восстановить. (int)
        :return: `True`, если аудиотрек успешно восстановлен, `False` в противном случае.
        """

        response = await self._VKReq("restore", {"owner_id": ownerId, "audio_id": trackId})

        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def createPlaylist(self, title: Union[str, int], description: Union[str, int] = None, photo: str = None, groupId: int = None, chatId: int = None) -> Union[int, dict, None]:
        """
        Создаёт плейлист в музыке пользователя или группы.

        Пример использования:\n
        result = await client.createPlaylist(title="prombl — npc", description="Release Date: December 24, 2021", "photo"="yourPhotoUrl", "groupId"="yourGroupId", chatId="yourChatId")\n
        print(result)

        :param title: название плейлиста. (str)
        :param description: описание плейлиста. (str, необязательно)
        :param photo: фото плейлиста. (str, необязательно, временно не работает)
        :param groupId: идентификатор группы, в которой необходимо создать плейлист. (int, необязательно)
        :param chatId: идентификатор чата, к которому привязать плейлист. (int, формат: `2000000000 + идентификатор чата`)
        :return: идентификатор плейлиста или словарь с ключами `ownerId` и `playlistId` (если создаётся плейлист, привязанный к чату), если плейлист успешно создан, `None` в противном случае.
        """

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        if chatId:
            method = "createChatPlaylist"

        else:
            method = "createPlaylist"

        playlist = await self._VKReq(method, {"title": title, "description": description, "owner_id": groupId})
        playlistId = playlist.get("id")

        if playlistId:
            if chatId:
                playlistOwnerId = playlist.get("owner_id")
                return {"ownerId": playlistOwnerId, "playlistId": playlistId}

            return playlistId

            """if photo:
                await self._editPlaylistPhoto(playlistId, photo, groupId)"""

        return playlist


    async def addPlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None) -> Union[int, None, dict]:
        """
        Добавляет плейлист в музыку пользователя или группы.

        Пример использования:\n
        result = await client.addPlaylist(playlistId=1, ownerId=-215973356, groupId="yourGroupId")\n
        print(result)

        :param ownerId: идентификатор владельца плейлиста (пользователь или группа). (int, по умолчанию текущий пользователь)
        :param playlistId: идентификатор плейлиста, который необходимо добавить. (int)
        :param groupId: идентификатор группы, в которую необходимо добавить плейлист. (int, необязательно)
        :return: идентификатор плейлиста, если плейлист успешно добавлен, `None` в противном случае.
        """

        if not ownerId:
            if not groupId:
                return None

            ownerId = (await self.getSelf()).get("id")

        response = await self._VKReq("followPlaylist", {"owner_id": ownerId, "playlist_id": playlistId, **({"group_id": groupId} if groupId else {})})
        return response.get("id") if not (isinstance(response, dict) and response.get("error")) else response


    async def removePlaylist(self, playlistId: int, groupId: int = None) -> Union[bool, dict]:
        """
        Удаляет плейлист из музыки пользователя или группы.

        Пример использования:\n
        result = await client.removePlaylist(playlistId="yourPlaylistId", groupId="yourGroupId")\n
        print(result)

        :param playlistId: идентификатор плейлиста, который необходимо удалить. (int)
        :param groupId: идентификатор группы, из которой необходимо удалить плейлист. (int, необязательно)
        :return: `True`, если плейлист успешно удалён, `False` в противном случае.
        """

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        response = await self._VKReq("deletePlaylist", {"owner_id": groupId, "playlist_id": playlistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else (response if response.get("error").get("code") != 17 else self._raiseError("playlistNotFound"))


    async def _editPlaylistPhoto(self, playlistId: int, photo: str = None, groupId: int = None) -> Union[bool, dict]:
        if photo is None:
            return False

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        params = {"playlist_id": playlistId, "owner_id": groupId}
        if photo:
            method = "setPlaylistCoverPhoto"
            params = {**params, **{"photo": photo}}

        else:
            method = "deletePlaylistCoverPhoto"

        response = await self._VKReq(method, params)
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def editPlaylist(self, playlistId: int, title: Union[str, int] = None, description: Union[str, int] = None, photo: str = None, groupId: int = None) -> Union[bool, dict]:
        """
        Изменяет информацию плейлиста, принадлежащего пользователю или группе.

        Пример использования:\n
        result = await client.editPlaylist(playlistId="yourPlaylistId", title="prombl — npc", description="Release Date: December 24, 2021", photo="yourPhotoUrl", groupId="yourGroupId")\n
        print(result)

        :param playlistId: идентификатор плейлиста, информацию которого необходимо изменить.
        :param title: новое название плейлиста. (Необязательно)
        :param description: новое описание плейлиста. (Необязательно)
        :param photo: новое фото плейлиста. (Необязательно)
        :param groupId: идентификатор группы, в которой находится плейлист. (Необязательно)
        :return: `True`, если информация плейлиста успешно обновлена, `False` в противном случае.
        """

        if not any((title, description is not None, photo is not None)):
            return False

        if not groupId:
            groupId = (await self.getSelf()).get("id")

        params = {
            "playlist_id": playlistId,
            "owner_id": groupId,
            **({"title": title} if title else {}),
            **({"description": description} if description is not None else {})
        }

        if not any(("title" in params, "description" in params, photo is not None)):
            return False

        else:
            if any(("title" in params, "description" in params)):
                response = await self._VKReq("editPlaylist", params)
                editInfoStatus = bool(response) if not (isinstance(response, dict) and response.get("error")) else response

            else:
                editInfoStatus = None

        if any((editInfoStatus is True, editInfoStatus is None)) and photo is not None:
            editPhotoStatus = await self._editPlaylistPhoto(playlistId, photo, groupId)
            if editInfoStatus is None:
                editInfoStatus = editPhotoStatus

        return editInfoStatus


    async def copyPlaylist(self, playlistId: int, ownerId: int = None, groupId: int = None, skipTitle: bool = False, skipDescription: bool = False, skipPhoto: bool = False) -> Union[int, None, dict]:
        """
        Копирует плейлист, принадлежий пользователю или группе в музыку пользователя или группы.

        Пример использования:\n
        result = await client.copyPlaylist(playlistId=1, ownerId=-215973356, groupId: "yourGroupId", skipTitle=True, skipDescription=True, skipPhoto=True)\n
        print(result)

        :param playlistId: идентификатор плейлиста, который необходимо скопировать. (int)
        :param ownerId: идентификатор владельца плейлиста (пользователь или группа). (int, по умолчанию текущий пользователь)
        :param groupId: идентификатор группы, в которую необходимо скопировать плейлист. (int, необязательно)
        :param skipTitle: флаг, указывающий, необходимо ли не копировать название плейлиста (если `True`, будут использованы текущие дата и время). (bool, по умолчанию `False`)
        :param skipDescription: флаг, указывающий, необходимо ли не копировать описание плейлиста. (bool, по умолчанию `False`)
        :param skipPhoto: флаг, указывающий, необходимо ли не копировать фото плейлиста. (bool, по умолчанию `False`)
        :return: идентификатор плейлиста, если плейлист успешно скопирован, `None` в противном случае.
        """

        if not ownerId:
            ownerId = (await self.getSelf()).get("id")

        playlist = await self.getPlaylist(playlistId, ownerId, True)

        if isinstance(playlist, dict):
            if playlist.get("error"):
                return playlist

            title = playlist.get("title")
            description = playlist.get("description")
            photo = playlist.get("photo")

        else:
            title = playlist.title
            description = playlist.description if hasattr(playlist, "description") else None
            photo = playlist.photo if hasattr(playlist, "photo") else None

        if skipTitle:
            title = datetime.utcnow().replace(tzinfo=pytz.utc).astimezone(moscowTz).strftime("%d.%m.%Y / %H:%M:%S")

        if skipDescription:
            description = None

        if skipPhoto:
            photo = None

        elif photo:
            _, photo = photo.popitem()

        newPlaylistId = await self.createPlaylist(title, description, photo, groupId)
        if isinstance(newPlaylistId, dict) and newPlaylistId.get("error"):
            return newPlaylistId

        tracks = (playlist.tracks if hasattr(playlist, "tracks") else None) if isinstance(playlist, Playlist) else playlist.get("tracks") if isinstance(playlist, dict) else None
        if not tracks:
            return newPlaylistId

        ownerIds = []
        trackIds = []
        for track in tracks[::-1]:
            ownerId, trackId = (track.ownerId, track.trackId) if isinstance(track, Track) else (track.get("ownerId"), track.get("trackId"))
            ownerIds.append(ownerId)
            trackIds.append(trackId)
        await self.add(ownerIds, trackIds, newPlaylistId, groupId)

        return newPlaylistId


    async def reorder(self, trackId: int, beforeTrackId: int = None, afterTrackId: int = None) -> Union[bool, dict]:
        """
        Изменяет порядок аудиотрека в музыке пользователя. Должен быть заполнен один из типов аргументов на выбор: `beforeTrackId` или `afterTrackId`.

        Пример использования для перемещения на место перед определённым треком:\n
        result = await client.reorder(trackId="yourTrackId", beforeTrackId="yourBeforeTrackId")\n
        print(result)

        Пример использования для перемещения на место после определённого трека:\n
        result = await client.reorder(trackId="yourTrackId", afterTrackId="yourAfterTrackId")\n
        print(result)

        :param trackId: идентификатор аудиотрека, порядок которого необходимо изменить. (int)
        :param beforeTrackId: идентификатор аудиотрека перед которым необходимо поместить аудиотрек. (int, необязательно)
        :param afterTrackId: идентификатор аудиотрека после которого необходимо поместить аудиотрек. (int, необязательно)
        :return: `True`, если порядок трека успешно изменён, `False` в противном случае.
        """

        if not any((beforeTrackId, afterTrackId)):
            return self._raiseError("trackReorderNeedsBeforeOrAfterArgument")

        if all((beforeTrackId, afterTrackId)):
            return self._raiseError("trackReorderNeedsOnlyBeforOrAfterNotBoth")

        response = await self._VKReq("reorder", {"audio_id": trackId, **({"before": beforeTrackId} if beforeTrackId else {"after": afterTrackId})})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def followArtist(self, artistId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки артиста.

        Пример использования:\n
        result = await client.followArtist(artistId=5696274288194638935)\n
        print(result)

        :param artistId: идентификатор артиста, на обновления которого необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки артиста, `False` в противном случае.
        """

        response = await self._VKReq("followArtist", {"artist_id": artistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def unfollowArtist(self, artistId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки артиста.

        Пример использования:\n
        result = await client.unfollowArtist(artistId=5696274288194638935)\n
        print(result)

        :param artistId: идентификатор артиста, от обновлений которого необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались от обновлений музыки артиста, `False` в противном случае.
        """

        response = await self._VKReq("unfollowArtist", {"artist_id": artistId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def followCurator(self, curatorId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки куратора.

        Пример использования:\n
        result = await client.followCurator(curatorId=28905875)\n
        print(result)

        :param curatorId: идентификатор куратора, на обновления которого необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки куратора, `False` в противном случае.
        """

        response = await self._VKReq("followCurator", {"curator_id": curatorId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def unfollowCurator(self, curatorId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки куратора.

        Пример использования:\n
        result = await client.unfollowCurator(curatorId=28905875)\n
        print(result)

        :param curatorId: идентификатор куратора, от обновлений которого необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались от обновлений музыки куратора, `False` в противном случае.
        """

        response = await self._VKReq("unfollowCurator", {"curator_id": curatorId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def followOwner(self, ownerId: int = None) -> Union[bool, dict]:
        """
        Подписывается на обновления музыки пользователя или группы.

        Пример использования:\n
        result = await client.followOwner(ownerId=-215973356)\n
        print(result)

        :param ownerId: идентификатор пользователя или группы, на обновления которого(ой) необходимо подписаться. (int)
        :return: `True`, если Вы успешно подписались на обновления музыки пользователя или группы, `False` в противном случае.
        """

        response = await self._VKReq("followOwner", {"owner_id": ownerId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def unfollowOwner(self, ownerId: int = None) -> Union[bool, dict]:
        """
        Отписывается от обновлений музыки пользователя или группы.

        Пример использования:\n
        result = await client.unfollowOwner(ownerId=-215973356)\n
        print(result)

        :param ownerId: идентификатор пользователя или группы, от обновлений которого(ой) необходимо отписаться. (int)
        :return: `True`, если Вы успешно отписались на обновлений музыки пользователя или группы, `False` в противном случае.
        """

        response = await self._VKReq("followOwner", {"owner_id": ownerId})
        return bool(response) if not (isinstance(response, dict) and response.get("error")) else response


    async def setBroadcast(self, ownerId: int = None, trackId: int = None, groupIds: Union[List[str], str] = None) -> Union[bool, dict]:
        """
        Устанавливает (удаляет) аудиотрек в (из) статус(а) пользователя или группы.

        Пример использования для установки аудиотрека в статус пользователя:\n
        result = await client.setBroadcast(ownerId=474499156, trackId=456637846)\n
        print(result)

        Пример использования для установки аудиотрека в статус группы:\n
        result = await client.setBroadcast(ownerId=474499156, trackId=456637846, groupdIds="yourGroupId")\n
        print(result)

        Пример использования для удаления аудиотрека из статуса пользователя:\n
        result = await client.setBroadcast()\n
        print(result)

        Пример использования для удаления аудиотрека из статуса группы:\n
        result = await client.setBroadcast(groupdIds="yourGroupId")\n
        print(result)

        :param ownerId: идентификатор владельца аудиотрека (пользователь или группа). (int, необязательно)
        :param trackId: идентификатор аудиотрека, который необходимо установить в статус. (int, необязательно)
        :param groupIds: идентификатор(ы) групп(ы), в (из) статус(а) которой необходимо установить (удалить) аудиотрек. (int, по умолчанию текущий пользователь)
        :return: `True`, если аудиотрек успешно установлен (удалён) в (из) статус(а), `False` в противном случае.
        """

        response = await self._VKReq("setBroadcast", {**({"audio": f"{ownerId}_{trackId}"} if ownerId and trackId else {}), **({"target_ids": groupIds} if groupIds else {})})
        return True if not (isinstance(response, dict) and response.get("error")) else response


from setuptools import setup, find_packages

setup(
    name="vkmusix",
    version="1.0.2",
    description="Библиотека для взаимодействия с VK Music. Документация: vkmusix.ru/docs.",
    long_description=open('README.md').read(),  # Чтение описания из файла README.md
    long_description_content_type='text/markdown',  # Указание формата README.md
    author="thswq",
    author_email="admin@vkmusix.ru",
    url="https://github.com/to4no4sv/vkmusix",
    packages=find_packages(),
    install_requires=[
        "pytz == 2024.1",
        "httpx == 0.27.0",
        "aiofiles == 23.2.1",
        "selenium == 4.21.0",
        "webdriver_manager == 4.0.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
